from __future__ import annotations

from .turbo_utils import standardize_y


def subsample_loglik(
    model,
    x,
    y,
    *,
    paramss: list,
    P: int = 10,
    rng,
) -> list:
    import numpy as np

    if x.ndim != 2:
        raise ValueError(x.shape)
    if y.ndim != 1:
        raise ValueError(y.shape)
    if x.shape[0] != y.shape[0]:
        raise ValueError((x.shape, y.shape))
    if P <= 0:
        raise ValueError(P)
    if len(paramss) == 0:
        raise ValueError("paramss must be non-empty")
    n = x.shape[0]
    if n == 0:
        return [0.0] * len(paramss)
    if len(model) <= 1:
        return [0.0] * len(paramss)
    P_actual = min(P, n)
    if P_actual == n:
        indices = np.arange(n, dtype=int)
    else:
        indices = rng.permutation(n)[:P_actual]
    x_selected = x[indices]
    y_selected = y[indices]
    if not np.isfinite(y_selected).all():
        return [0.0] * len(paramss)
    post_batch = model.batch_posterior(x_selected, paramss, exclude_nearest=True)
    mu_batch = post_batch.mu
    se_batch = post_batch.se
    if mu_batch.shape[2] == 1:
        mu_batch = mu_batch[:, :, 0]
        se_batch = se_batch[:, :, 0]
    num_params = len(paramss)
    if mu_batch.shape != (num_params, P_actual) or se_batch.shape != (
        num_params,
        P_actual,
    ):
        raise ValueError((mu_batch.shape, se_batch.shape, (num_params, P_actual)))
    _, y_std = standardize_y(y)
    y_scaled = y_selected / y_std
    mu_scaled = mu_batch / y_std
    se_scaled = se_batch / y_std
    result = []
    for i in range(num_params):
        mu_i = mu_scaled[i]
        se_i = se_scaled[i]
        if not np.isfinite(mu_i).all() or not np.isfinite(se_i).all():
            result.append(0.0)
            continue
        if np.any(se_i <= 0.0):
            result.append(0.0)
            continue
        diff = y_scaled - mu_i
        var_scaled = se_i**2
        log_term = np.log(2.0 * np.pi * var_scaled)
        quad = diff**2 / var_scaled
        loglik = -0.5 * np.sum(log_term + quad)
        if not np.isfinite(loglik):
            result.append(0.0)
            continue
        result.append(float(loglik))
    return result


def enn_fit(
    model,
    *,
    k: int | None = None,
    num_fit_candidates: int,
    num_fit_samples: int = 10,
    rng,
):
    from .enn_params import ENNParams

    if k is not None:
        return ENNParams(k=k, var_scale=1.0)
    train_x = model.train_x
    train_y = model.train_y
    train_yvar = model.train_yvar
    if train_y.shape[1] != 1 or train_yvar.shape[1] != 1:
        raise ValueError((train_y.shape, train_yvar.shape))
    y = train_y[:, 0]
    k_candidates = [3, 4, 5, 6, 7, 8, 9, 10, 20, 30, 50, 100]
    k_values = rng.choice(k_candidates, size=num_fit_candidates, replace=True).tolist()
    var_scale_log_min = -2.0
    var_scale_log_max = 2.0
    var_scale_log_values = rng.uniform(
        var_scale_log_min, var_scale_log_max, size=num_fit_candidates
    )
    var_scale_values = (10**var_scale_log_values).tolist()
    paramss = [
        ENNParams(k=k_val, var_scale=var_scale_val)
        for k_val, var_scale_val in zip(k_values, var_scale_values)
    ]
    if len(paramss) == 0:
        return ENNParams(k=10, var_scale=1.0)
    logliks = subsample_loglik(
        model, train_x, y, paramss=paramss, P=num_fit_samples, rng=rng
    )
    best_idx: int | None = None
    best_mll: float | None = None
    for i, loglik in enumerate(logliks):
        if best_mll is None or loglik > best_mll:
            best_mll = loglik
            best_idx = i
    if best_idx is None:
        return ENNParams(k=k_values[0], var_scale=float(var_scale_values[0]))
    return paramss[best_idx]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np

    from .enn_params import ENNParams


class EpistemicNearestNeighbors:
    def __init__(
        self,
        train_x,
        train_y,
        train_yvar,
        hnsw_threshold: int | None = None,
        hnsw_M: int = 32,
    ) -> None:
        import numpy as np

        if train_x.ndim != 2:
            raise ValueError(train_x.shape)
        if train_y.ndim != 2 or train_yvar.ndim != 2:
            raise ValueError((train_y.shape, train_yvar.shape))
        if train_x.shape[0] != train_y.shape[0] or train_y.shape != train_yvar.shape:
            raise ValueError((train_x.shape, train_y.shape, train_yvar.shape))
        self._train_x = np.asarray(train_x, dtype=float)
        self._train_y = np.asarray(train_y, dtype=float)
        self._train_yvar = np.asarray(train_yvar, dtype=float)
        self._num_obs, self._num_dim = self._train_x.shape
        _, self._num_metrics = self._train_y.shape
        self._eps_var = 1e-9
        self._x_scale = np.std(self._train_x, axis=0).astype(float)
        self._x_scale = np.maximum(self._x_scale, 1e-6)
        if len(self._train_y) < 2:
            self._y_scale = np.ones(shape=(1, self._num_metrics), dtype=float)
        else:
            self._y_scale = np.std(self._train_y, axis=0, keepdims=True).astype(float)

        self._index = None
        self._hnsw_threshold = hnsw_threshold
        self._hnsw_M = int(hnsw_M)
        self._build_index()

    @property
    def train_x(self) -> np.ndarray:
        return self._train_x

    @property
    def train_y(self) -> np.ndarray:
        return self._train_y

    @property
    def train_yvar(self) -> np.ndarray:
        return self._train_yvar

    @property
    def num_outputs(self) -> int:
        return self._num_metrics

    def __len__(self) -> int:
        return self._num_obs

    def _build_index(self) -> None:
        import faiss
        import numpy as np

        if self._num_obs == 0:
            return
        x_scaled = self._train_x / self._x_scale
        x_scaled = x_scaled.astype(np.float32, copy=False)
        if self._hnsw_threshold is not None and self._num_obs >= self._hnsw_threshold:
            index = faiss.IndexHNSWFlat(self._num_dim, self._hnsw_M)
        else:
            index = faiss.IndexFlatL2(self._num_dim)
        index.add(x_scaled)
        self._index = index

    def posterior(
        self,
        x,
        *,
        params: ENNParams,
        exclude_nearest: bool = False,
    ):
        from .enn_normal import ENNNormal

        post_batch = self.batch_posterior(x, [params], exclude_nearest=exclude_nearest)
        mu = post_batch.mu[0]
        se = post_batch.se[0]
        return ENNNormal(mu, se)

    def batch_posterior(
        self,
        x,
        paramss: list[ENNParams],
        *,
        exclude_nearest: bool = False,
    ):
        import numpy as np

        from .enn_normal import ENNNormal

        if x.ndim != 2:
            raise ValueError(x.shape)
        if x.shape[1] != self._num_dim:
            raise ValueError(x.shape)
        if len(paramss) == 0:
            raise ValueError("paramss must be non-empty")
        batch_size = x.shape[0]
        num_params = len(paramss)
        if len(self) == 0:
            mu = np.zeros((num_params, batch_size, self._num_metrics), dtype=float)
            se = np.ones((num_params, batch_size, self._num_metrics), dtype=float)
            return ENNNormal(mu, se)
        max_k = max(params.k for params in paramss)
        if exclude_nearest:
            if len(self) <= 1:
                raise ValueError(len(self))
            search_k = int(min(max_k + 1, len(self)))
        else:
            search_k = int(min(max_k, len(self)))
        x_scaled = x / self._x_scale
        x_scaled = x_scaled.astype(np.float32, copy=False)
        if self._index is None:
            raise RuntimeError("index is not initialized")
        dist2s_full, idx_full = self._index.search(x_scaled, search_k)
        dist2s_full = dist2s_full.astype(float)
        idx_full = idx_full.astype(int)
        if exclude_nearest:
            dist2s_full = dist2s_full[:, 1:]
            idx_full = idx_full[:, 1:]
        mu_all = np.zeros((num_params, batch_size, self._num_metrics), dtype=float)
        se_all = np.zeros((num_params, batch_size, self._num_metrics), dtype=float)
        for i, params in enumerate(paramss):
            k = min(params.k, search_k)
            if k == 0:
                mu_all[i] = np.zeros((batch_size, self._num_metrics), dtype=float)
                se_all[i] = np.ones((batch_size, self._num_metrics), dtype=float)
                continue
            dist2s = dist2s_full[:, :k]
            idx = idx_full[:, :k]
            y_neighbors = self._train_y[idx]
            yvar_neighbors = self._train_yvar[idx]
            dist2s_expanded = dist2s[..., np.newaxis]
            var_component = (
                params.var_scale * dist2s_expanded + yvar_neighbors / self._y_scale**2
            )
            w = 1.0 / (self._eps_var + var_component)
            norm = np.sum(w, axis=1)
            mu_all[i] = np.sum(w * y_neighbors, axis=1) / norm
            epistemic_var = 1.0 / norm
            noise_var = np.sum(w * yvar_neighbors, axis=1) / norm
            vvar = epistemic_var + noise_var
            vvar = np.maximum(vvar, self._eps_var)
            se_all[i] = np.sqrt(vvar) * self._y_scale
        return ENNNormal(mu_all, se_all)



from __future__ import annotations

import conftest
import pytest

from enn.turbo_mode import TurboMode
from enn.turbo_utils import to_unit


def _run_bo(mode: TurboMode, num_steps: int = 15) -> float:
    import numpy as np

    from enn import Turbo

    bounds = np.array([[-1.0, 1.0], [-1.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    opt = Turbo(bounds=bounds, mode=mode, num_arms=4, rng=rng)
    best = -np.inf
    for _ in range(num_steps):
        x = opt.ask(num_arms=4)
        y = conftest.sphere_objective(x)
        opt.tell(x, y)
        best = max(best, float(np.max(y)))
    return best


def test_turbo_zero_ask_tell_and_shape():
    import numpy as np

    from enn import Turbo

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    opt = Turbo(
        bounds=bounds,
        mode=TurboMode.TURBO_ZERO,
        num_arms=3,
        rng=rng,
    )
    x0 = opt.ask(num_arms=3)
    assert x0.shape == (3, 2)
    assert np.all(x0 >= 0.0) and np.all(x0 <= 1.0)
    y0 = conftest.sphere_objective(x0)
    opt.tell(x0, y0)
    x1 = opt.ask(num_arms=3)
    assert x1.shape == (3, 2)


def test_turbo_one_improves_on_sphere():
    best = _run_bo(TurboMode.TURBO_ONE, num_steps=12)
    assert best > -0.5


def test_turbo_zero_reasonable_on_sphere():
    best = _run_bo(TurboMode.TURBO_ZERO, num_steps=12)
    assert best > -1.5


def test_turbo_enn_uses_enn_and_is_reasonable():
    best = _run_bo(TurboMode.TURBO_ENN, num_steps=12)
    assert best > -1.5


def test_turbo_enn_with_k_none_fits_hyperparameters():
    import numpy as np

    from enn.turbo_mode import TurboMode
    from enn.turbo_optimizer import TurboOptimizer

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    opt = TurboOptimizer(
        bounds=bounds,
        mode=TurboMode.TURBO_ENN,
        num_arms=3,
        rng=rng,
        k=None,
    )
    x0 = opt.ask(num_arms=3)
    assert x0.shape == (3, 2)
    assert np.all(x0 >= 0.0) and np.all(x0 <= 1.0)
    y0 = -np.sum(x0**2, axis=1)
    opt.tell(x0, y0)
    x1 = opt.ask(num_arms=3)
    assert x1.shape == (3, 2)
    assert np.all(x1 >= 0.0) and np.all(x1 <= 1.0)


def test_turbo_optimizer_with_trailing_obs():
    import numpy as np

    from enn.turbo_mode import TurboMode
    from enn.turbo_optimizer import TurboOptimizer

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)

    for mode in [TurboMode.TURBO_ONE, TurboMode.TURBO_ENN]:
        opt = TurboOptimizer(
            bounds=bounds,
            mode=mode,
            num_arms=2,
            rng=rng,
            trailing_obs=5,
        )
        for i in range(10):
            x = opt.ask(num_arms=2)
            assert x.shape == (2, 2)
            y = -np.sum(x**2, axis=1)
            opt.tell(x, y)
        assert len(opt._x_obs_list) == 20
        assert len(opt._y_obs_list) == 20
        x_final = opt.ask(num_arms=2)
        assert x_final.shape == (2, 2)


def test_latin_hypercube_stratification_and_bounds():
    import numpy as np

    from enn.turbo_utils import latin_hypercube

    rng = np.random.default_rng(0)
    n = 8
    d = 3
    x = latin_hypercube(n, d, rng=rng)
    assert x.shape == (n, d)
    assert np.all(x >= 0.0) and np.all(x <= 1.0)
    for j in range(d):
        xs = np.sort(x[:, j])
        for k in range(n):
            lo = k / n
            hi = (k + 1) / n
            in_bin = (xs >= lo) & (xs <= hi + 1e-8)
            assert np.any(in_bin)


def test_argmax_random_tie_uses_rng_and_is_deterministic():
    import numpy as np

    from enn.turbo_utils import argmax_random_tie

    values = np.array([1.0, 2.0, 2.0, 0.0], dtype=float)
    rng = np.random.default_rng(0)
    idx1 = argmax_random_tie(values, rng=rng)
    assert idx1 in (1, 2)
    rng = np.random.default_rng(0)
    idx2 = argmax_random_tie(values, rng=rng)
    assert idx1 == idx2


def test_pareto_front_simple_case():
    import numpy as np

    from enn.turbo_utils import pareto_front

    mu = np.array([1.0, 0.9, 0.8], dtype=float)
    se = np.array([1.0, 0.5, 0.7], dtype=float)
    mask = pareto_front(mu, se)
    assert mask.dtype == bool
    assert mask.shape == mu.shape
    assert mask[0]
    assert mask[1]
    assert not mask[2]


def test_trust_region_state_update_and_restart_and_bounds():
    import numpy as np

    from enn.trust_region_state import TrustRegionState

    state = TrustRegionState(num_dim=2, num_arms=2)
    values = []
    for v in [0.0, 1.0, 2.0]:
        values.append(v)
        state.update(np.array(values, dtype=float))
    x_center = np.zeros((1, 2), dtype=float)
    lb, ub = state.create_bounds(x_center)
    assert lb.shape == (1, 2)
    assert ub.shape == (1, 2)
    state.length = state.length_min / 2.0
    assert state.needs_restart()
    state.restart()
    assert state.length == state.length_init


@pytest.mark.parametrize(
    "mode", [TurboMode.TURBO_ZERO, TurboMode.TURBO_ONE, TurboMode.TURBO_ENN]
)
def test_turbo_behavior_independent_of_affine_x(mode: TurboMode) -> None:
    import numpy as np

    from enn import Turbo

    bounds1 = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    bounds2 = np.array([[2.0, 4.0], [-3.0, 1.0]], dtype=float)
    num_arms = 3
    num_steps = 8
    rng1 = np.random.default_rng(0)
    rng2 = np.random.default_rng(0)
    opt1 = Turbo(
        bounds=bounds1,
        mode=mode,
        num_arms=num_arms,
        rng=rng1,
    )
    opt2 = Turbo(
        bounds=bounds2,
        mode=mode,
        num_arms=num_arms,
        rng=rng2,
    )

    for _ in range(num_steps):
        x1 = opt1.ask(num_arms=num_arms)
        x2 = opt2.ask(num_arms=num_arms)
        u1 = to_unit(x1, bounds1)
        u2 = to_unit(x2, bounds2)
        assert np.allclose(u1, u2)
        z1 = 2.0 * u1 - 1.0
        z2 = 2.0 * u2 - 1.0
        y1 = conftest.sphere_objective(z1)
        y2 = conftest.sphere_objective(z2)
        assert np.allclose(y1, y2)
        opt1.tell(x1, y1)
        opt2.tell(x2, y2)


@pytest.mark.parametrize("mode", [TurboMode.TURBO_ZERO, TurboMode.TURBO_ENN])
def test_turbo_behavior_independent_of_affine_y(mode: TurboMode) -> None:
    import numpy as np

    from enn import Turbo

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    num_arms = 3
    num_steps = 8

    def run_with_transform(scale: float, shift: float) -> np.ndarray:
        rng = np.random.default_rng(0)
        opt = Turbo(
            bounds=bounds,
            mode=mode,
            num_arms=num_arms,
            rng=rng,
        )
        unit_trajectory = []
        for _ in range(num_steps):
            x = opt.ask(num_arms=num_arms)
            # With these bounds, x already lives in unit space.
            u = x.copy()
            z = 2.0 * u - 1.0
            base_y = conftest.sphere_objective(z)
            y = scale * base_y + shift
            opt.tell(x, y)
            unit_trajectory.append(u)
        return np.stack(unit_trajectory, axis=0)

    traj_base = run_with_transform(scale=1.0, shift=0.0)
    traj_affine = run_with_transform(scale=2.0, shift=0.5)
    # The sequence of unit-space query points should be invariant to affine
    # rescalings (scale and center) of the observed y values.
    # Note: TURBO_ONE is excluded because it uses a relative threshold
    # (1e-3 * abs(best_value)) for improvement detection, which breaks affine
    # invariance by design (matching the reference implementation).
    assert np.allclose(traj_base, traj_affine)


def test_sobol_perturb_np_shape_and_bounds():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import sobol_perturb_np

    num_candidates = 10
    num_dim = 3
    x_center = np.array([0.5, 0.5, 0.5], dtype=float)
    lb = np.array([0.0, 0.0, 0.0], dtype=float)
    ub = np.array([1.0, 1.0, 1.0], dtype=float)
    mask = np.ones((num_candidates, num_dim), dtype=bool)
    sobol_engine = qmc.Sobol(d=num_dim, scramble=True, seed=0)
    candidates = sobol_perturb_np(
        x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine
    )
    assert candidates.shape == (num_candidates, num_dim)
    assert np.all(candidates >= lb)
    assert np.all(candidates <= ub)


def test_sobol_perturb_np_mask_application():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import sobol_perturb_np

    num_candidates = 5
    num_dim = 3
    x_center = np.array([0.5, 0.5, 0.5], dtype=float)
    lb = np.array([0.0, 0.0, 0.0], dtype=float)
    ub = np.array([1.0, 1.0, 1.0], dtype=float)
    mask = np.zeros((num_candidates, num_dim), dtype=bool)
    mask[:, 0] = True
    mask[0, 1] = True
    sobol_engine = qmc.Sobol(d=num_dim, scramble=True, seed=0)
    candidates = sobol_perturb_np(
        x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine
    )
    assert candidates.shape == (num_candidates, num_dim)
    for i in range(num_candidates):
        for j in range(num_dim):
            if mask[i, j]:
                assert candidates[i, j] != x_center[j]
                assert lb[j] <= candidates[i, j] <= ub[j]
            else:
                assert candidates[i, j] == x_center[j]


def test_sobol_perturb_np_deterministic():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import sobol_perturb_np

    num_candidates = 8
    num_dim = 2
    x_center = np.array([0.5, 0.5], dtype=float)
    lb = np.array([0.0, 0.0], dtype=float)
    ub = np.array([1.0, 1.0], dtype=float)
    mask = np.ones((num_candidates, num_dim), dtype=bool)
    sobol_engine1 = qmc.Sobol(d=num_dim, scramble=True, seed=42)
    sobol_engine2 = qmc.Sobol(d=num_dim, scramble=True, seed=42)
    candidates1 = sobol_perturb_np(
        x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine1
    )
    candidates2 = sobol_perturb_np(
        x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine2
    )
    assert np.allclose(candidates1, candidates2)


def test_raasp_shape_and_bounds():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import raasp

    num_candidates = 10
    num_dim = 3
    x_center = np.array([0.5, 0.5, 0.5], dtype=float)
    lb = np.array([0.0, 0.0, 0.0], dtype=float)
    ub = np.array([1.0, 1.0, 1.0], dtype=float)
    rng = np.random.default_rng(0)
    sobol_engine = qmc.Sobol(d=num_dim, scramble=True, seed=0)
    candidates = raasp(
        x_center,
        lb,
        ub,
        num_candidates,
        num_pert=20,
        rng=rng,
        sobol_engine=sobol_engine,
    )
    assert candidates.shape == (num_candidates, num_dim)
    assert np.all(candidates >= lb)
    assert np.all(candidates <= ub)


def test_raasp_at_least_one_dimension_perturbed():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import raasp

    num_candidates = 20
    num_dim = 5
    x_center = np.array([0.5] * num_dim, dtype=float)
    lb = np.array([0.0] * num_dim, dtype=float)
    ub = np.array([1.0] * num_dim, dtype=float)
    rng = np.random.default_rng(0)
    sobol_engine = qmc.Sobol(d=num_dim, scramble=True, seed=0)
    candidates = raasp(
        x_center,
        lb,
        ub,
        num_candidates,
        num_pert=20,
        rng=rng,
        sobol_engine=sobol_engine,
    )
    for i in range(num_candidates):
        diff = np.abs(candidates[i] - x_center)
        assert np.any(diff > 1e-10)


def test_raasp_deterministic():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import raasp

    num_candidates = 8
    num_dim = 2
    x_center = np.array([0.5, 0.5], dtype=float)
    lb = np.array([0.0, 0.0], dtype=float)
    ub = np.array([1.0, 1.0], dtype=float)
    rng1 = np.random.default_rng(42)
    rng2 = np.random.default_rng(42)
    sobol_engine1 = qmc.Sobol(d=num_dim, scramble=True, seed=0)
    sobol_engine2 = qmc.Sobol(d=num_dim, scramble=True, seed=0)
    candidates1 = raasp(
        x_center,
        lb,
        ub,
        num_candidates,
        num_pert=20,
        rng=rng1,
        sobol_engine=sobol_engine1,
    )
    candidates2 = raasp(
        x_center,
        lb,
        ub,
        num_candidates,
        num_pert=20,
        rng=rng2,
        sobol_engine=sobol_engine2,
    )
    assert np.allclose(candidates1, candidates2)


def test_raasp_probability_scaling():
    import numpy as np
    from scipy.stats import qmc

    from enn.turbo_utils import raasp

    num_candidates = 100
    num_dim_low = 5
    num_dim_high = 100
    x_center_low = np.array([0.5] * num_dim_low, dtype=float)
    x_center_high = np.array([0.5] * num_dim_high, dtype=float)
    lb_low = np.array([0.0] * num_dim_low, dtype=float)
    ub_low = np.array([1.0] * num_dim_low, dtype=float)
    lb_high = np.array([0.0] * num_dim_high, dtype=float)
    ub_high = np.array([1.0] * num_dim_high, dtype=float)
    rng = np.random.default_rng(0)
    sobol_engine_low = qmc.Sobol(d=num_dim_low, scramble=True, seed=0)
    sobol_engine_high = qmc.Sobol(d=num_dim_high, scramble=True, seed=0)
    candidates_low = raasp(
        x_center_low,
        lb_low,
        ub_low,
        num_candidates,
        num_pert=20,
        rng=rng,
        sobol_engine=sobol_engine_low,
    )
    rng = np.random.default_rng(0)
    candidates_high = raasp(
        x_center_high,
        lb_high,
        ub_high,
        num_candidates,
        num_pert=20,
        rng=rng,
        sobol_engine=sobol_engine_high,
    )
    diff_low = np.sum(np.abs(candidates_low - x_center_low) > 1e-10, axis=1)
    diff_high = np.sum(np.abs(candidates_high - x_center_high) > 1e-10, axis=1)
    mean_perturbed_low = np.mean(diff_low) / num_dim_low
    mean_perturbed_high = np.mean(diff_high) / num_dim_high
    assert mean_perturbed_low > mean_perturbed_high


def test_to_unit_and_from_unit_roundtrip():
    import numpy as np

    from enn.turbo_utils import from_unit

    bounds = np.array([[0.0, 2.0], [-1.0, 1.0], [5.0, 10.0]], dtype=float)
    x_original = np.array([[1.0, 0.0, 7.5], [0.5, -0.5, 8.0]], dtype=float)
    x_unit = to_unit(x_original, bounds)
    assert x_unit.shape == x_original.shape
    assert np.all(x_unit >= 0.0) and np.all(x_unit <= 1.0)
    x_roundtrip = from_unit(x_unit, bounds)
    assert np.allclose(x_original, x_roundtrip)


def test_to_unit_bounds_validation():
    import numpy as np

    bounds_invalid = np.array([[1.0, 0.0]], dtype=float)
    x = np.array([[0.5]], dtype=float)
    with pytest.raises(ValueError):
        to_unit(x, bounds_invalid)


def test_select_uniform_shape_and_uniformity():
    import numpy as np

    from enn.proposal import select_uniform

    num_candidates = 100
    num_dim = 3
    num_arms = 10
    x_cand = np.random.default_rng(0).random((num_candidates, num_dim))
    bounds = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    from_unit_fn = conftest.make_from_unit_fn(bounds)

    selected = select_uniform(x_cand, num_arms, num_dim, rng, from_unit_fn)
    assert selected.shape == (num_arms, num_dim)
    assert len(np.unique([tuple(row) for row in selected], axis=0)) == num_arms


def test_select_uniform_validation():
    import numpy as np

    from enn.proposal import select_uniform

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    from_unit_fn = conftest.make_from_unit_fn(bounds)

    x_cand_wrong_dim = np.random.default_rng(0).random((10, 3))
    with pytest.raises(ValueError):
        select_uniform(x_cand_wrong_dim, 5, 2, rng, from_unit_fn)

    x_cand_too_few = np.random.default_rng(0).random((3, 2))
    with pytest.raises(ValueError):
        select_uniform(x_cand_too_few, 5, 2, rng, from_unit_fn)


def test_select_enn_pareto_uses_enn_and_selects_pareto():
    import numpy as np

    from enn.proposal import select_enn_pareto

    num_candidates = 50
    num_dim = 2
    num_arms = 5
    x_cand = np.random.default_rng(0).random((num_candidates, num_dim))
    x_obs = np.random.default_rng(1).random((20, num_dim))
    y_obs = (x_obs.sum(axis=1)).tolist()
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    from_unit_fn = conftest.make_from_unit_fn(bounds)
    fallback_fn = conftest.make_fallback_fn(bounds, rng)

    selected = select_enn_pareto(
        x_cand,
        num_arms,
        x_obs.tolist(),
        y_obs,
        k=5,
        var_scale=1.0,
        hnsw_threshold=None,
        rng=rng,
        fallback_fn=fallback_fn,
        from_unit_fn=from_unit_fn,
    )
    assert selected.shape == (num_arms, num_dim)
    assert np.all(selected >= bounds[:, 0]) and np.all(selected <= bounds[:, 1])


def test_select_enn_pareto_fallback_on_empty_observations():
    import numpy as np

    from enn.proposal import select_enn_pareto

    num_candidates = 20
    num_dim = 2
    num_arms = 3
    x_cand = np.random.default_rng(0).random((num_candidates, num_dim))
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    from_unit_fn = conftest.make_from_unit_fn(bounds)

    fallback_called = False

    def fallback_fn(x, n):
        nonlocal fallback_called
        fallback_called = True
        idx = rng.choice(x.shape[0], size=n, replace=False)
        return from_unit_fn(x[idx])

    selected = select_enn_pareto(
        x_cand,
        num_arms,
        [],
        [],
        k=5,
        var_scale=1.0,
        hnsw_threshold=None,
        rng=rng,
        fallback_fn=fallback_fn,
        from_unit_fn=from_unit_fn,
    )
    assert fallback_called
    assert selected.shape == (num_arms, num_dim)


def test_select_enn_pareto_with_k_none_fits_model():
    import numpy as np

    from enn.proposal import select_enn_pareto

    num_candidates = 50
    num_dim = 2
    num_arms = 5
    x_cand = np.random.default_rng(0).random((num_candidates, num_dim))
    x_obs = np.random.default_rng(1).random((30, num_dim))
    y_obs = (x_obs.sum(axis=1)).tolist()
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    from_unit_fn = conftest.make_from_unit_fn(bounds)
    fallback_fn = conftest.make_fallback_fn(bounds, rng)

    selected = select_enn_pareto(
        x_cand,
        num_arms,
        x_obs.tolist(),
        y_obs,
        k=None,
        var_scale=1.0,
        hnsw_threshold=None,
        rng=rng,
        fallback_fn=fallback_fn,
        from_unit_fn=from_unit_fn,
    )
    assert selected.shape == (num_arms, num_dim)
    assert np.all(selected >= bounds[:, 0]) and np.all(selected <= bounds[:, 1])


def test_select_gp_thompson_uses_gp_and_returns_correct_shape():
    import numpy as np

    from enn.proposal import select_gp_thompson

    num_candidates = 30
    num_dim = 2
    num_arms = 5
    x_cand = np.random.default_rng(0).random((num_candidates, num_dim))
    x_obs = np.random.default_rng(1).random((15, num_dim))
    y_obs = (x_obs.sum(axis=1)).tolist()
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    from_unit_fn = conftest.make_from_unit_fn(bounds)
    select_sobol_fn = conftest.make_select_sobol_fn(bounds, rng)

    selected, new_mean, new_std, _ = select_gp_thompson(
        x_cand,
        num_arms,
        x_obs.tolist(),
        y_obs,
        num_dim,
        gp_num_steps=20,
        rng=rng,
        gp_y_mean=0.0,
        gp_y_std=1.0,
        select_sobol_fn=select_sobol_fn,
        from_unit_fn=from_unit_fn,
    )
    assert selected.shape == (num_arms, num_dim)
    assert isinstance(new_mean, float)
    assert isinstance(new_std, float)
    assert new_std > 0.0
    assert np.all(selected >= bounds[:, 0]) and np.all(selected <= bounds[:, 1])


def test_select_gp_thompson_fallback_on_empty_observations():
    import numpy as np

    from enn.proposal import select_gp_thompson

    num_candidates = 20
    num_dim = 2
    num_arms = 3
    x_cand = np.random.default_rng(0).random((num_candidates, num_dim))
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    from_unit_fn = conftest.make_from_unit_fn(bounds)

    fallback_called = False

    def select_sobol_fn(x, n):
        nonlocal fallback_called
        fallback_called = True
        idx = rng.choice(x.shape[0], size=n, replace=False)
        return from_unit_fn(x[idx])

    selected, mean, std, _ = select_gp_thompson(
        x_cand,
        num_arms,
        [],
        [],
        num_dim,
        gp_num_steps=20,
        rng=rng,
        gp_y_mean=0.0,
        gp_y_std=1.0,
        select_sobol_fn=select_sobol_fn,
        from_unit_fn=from_unit_fn,
    )
    assert fallback_called
    assert selected.shape == (num_arms, num_dim)
    assert mean == 0.0
    assert std == 1.0


def test_fit_gp_returns_model_with_valid_data():
    import numpy as np

    from enn.turbo_utils import fit_gp

    num_obs = 20
    num_dim = 3
    x_obs = np.random.default_rng(0).random((num_obs, num_dim))
    y_obs = (
        x_obs.sum(axis=1) + 0.1 * np.random.default_rng(1).standard_normal(num_obs)
    ).tolist()
    model, likelihood, y_mean, y_std = fit_gp(
        x_obs.tolist(), y_obs, num_dim, num_steps=10
    )
    assert model is not None
    assert likelihood is not None
    assert isinstance(y_mean, float)
    assert isinstance(y_std, float)
    assert y_std > 0.0


def test_fit_gp_returns_none_with_insufficient_data():
    import numpy as np

    from enn.turbo_utils import fit_gp

    num_dim = 2
    model_empty, likelihood_empty, mean_empty, std_empty = fit_gp(
        [], [], num_dim, num_steps=10
    )
    assert model_empty is None
    assert likelihood_empty is None
    assert mean_empty == 0.0
    assert std_empty == 1.0

    x_single = np.random.default_rng(0).random((1, num_dim))
    y_single = [1.0]
    model_single, likelihood_single, mean_single, std_single = fit_gp(
        x_single.tolist(), y_single, num_dim, num_steps=10
    )
    assert model_single is None
    assert likelihood_single is None
    assert isinstance(mean_single, float)
    assert std_single == 1.0



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .proposal import select_enn_pareto, select_gp_thompson, select_uniform
from .turbo_utils import argmax_random_tie, from_unit, latin_hypercube, raasp, to_unit

if TYPE_CHECKING:
    import numpy as np

    from .turbo_mode import TurboMode


class TurboOptimizer:
    def __init__(
        self,
        bounds,
        mode,
        num_arms: int,
        *,
        num_candidates: Optional[int] = None,
        rng,
        hnsw_threshold: Optional[int] = None,
        k: Optional[int] = None,
        var_scale: float = 1.0,
        trailing_obs: Optional[int] = None,
        num_init: Optional[int] = None,
    ) -> None:
        import numpy as np
        from scipy.stats import qmc

        from .trust_region_state import TrustRegionState

        if bounds.ndim != 2 or bounds.shape[1] != 2:
            raise ValueError(bounds.shape)
        self._bounds = np.asarray(bounds, dtype=float)
        self._num_dim = self._bounds.shape[0]
        self._mode = mode
        tr_num_arms = int(num_arms)
        if tr_num_arms <= 0:
            raise ValueError(tr_num_arms)
        if num_candidates is None:
            num_candidates = min(5000, 100 * self._num_dim)
        from .turbo_mode import TurboMode

        if mode == TurboMode.TURBO_ENN:
            num_candidates = max(num_candidates, 10 * tr_num_arms)
        self._num_candidates = int(num_candidates)
        if self._num_candidates <= 0:
            raise ValueError(self._num_candidates)
        self._rng = rng
        sobol_seed = int(self._rng.integers(1_000_000))
        self._sobol_engine = qmc.Sobol(d=self._num_dim, scramble=True, seed=sobol_seed)
        self._x_obs_list: list = []
        self._y_obs_list: list = []

        if mode == TurboMode.TURBO_ONE or mode == TurboMode.TURBO_ENN:
            self._x_tr_list: list = []
            self._y_tr_list: list = []
        else:
            self._x_tr_list = None
            self._y_tr_list = None
        self._tr_state = TrustRegionState(num_dim=self._num_dim, num_arms=tr_num_arms)
        self._gp_y_mean: float = 0.0
        self._gp_y_std: float = 1.0
        self._gp_num_steps: int = 50
        self._hnsw_threshold = hnsw_threshold
        if k is not None:
            k_val = int(k)
            if k_val < 3:
                raise ValueError(f"k must be >= 3, got {k_val}")
            self._k = k_val
        else:
            self._k = None
        var_scale_val = float(var_scale)
        if var_scale_val <= 0.0:
            raise ValueError(f"var_scale must be > 0, got {var_scale_val}")
        self._var_scale = var_scale_val
        if trailing_obs is not None:
            trailing_obs_val = int(trailing_obs)
            if trailing_obs_val <= 0:
                raise ValueError(f"trailing_obs must be > 0, got {trailing_obs_val}")
            self._trailing_obs = trailing_obs_val
        else:
            self._trailing_obs = None
        if num_init is None:
            num_init = 2 * self._num_dim
        num_init_val = int(num_init)
        if num_init_val <= 0:
            raise ValueError(f"num_init must be > 0, got {num_init_val}")
        self._num_init = num_init_val
        self._init_lhd = from_unit(
            latin_hypercube(self._num_init, self._num_dim, rng=self._rng),
            self._bounds,
        )
        self._init_idx = 0

    @property
    def num_dim(self) -> int:
        return self._num_dim

    @property
    def mode(self) -> TurboMode:
        return self._mode

    def ask(self, num_arms: int) -> np.ndarray:
        from .turbo_mode import TurboMode

        num_arms = int(num_arms)
        if num_arms <= 0:
            raise ValueError(num_arms)
        if self._mode == TurboMode.LHD_ONLY:
            return self._draw_initial(num_arms)
        if (
            self._mode == TurboMode.TURBO_ONE
            and self._x_tr_list is not None
            and len(self._x_tr_list) == 0
        ):
            return self._get_init_lhd_points(num_arms)
        if self._init_idx < self._num_init:
            if len(self._x_obs_list) == 0:
                fallback_fn = None
            else:
                fallback_fn = self._ask_normal
            return self._get_init_lhd_points(num_arms, fallback_fn=fallback_fn)
        if len(self._x_obs_list) == 0:
            return self._draw_initial(num_arms)
        return self._ask_normal(num_arms)

    def _ask_normal(self, num_arms: int) -> np.ndarray:
        from .turbo_mode import TurboMode

        if self._tr_state.needs_restart():
            self._tr_state.restart()
            if self._mode == TurboMode.TURBO_ONE or self._mode == TurboMode.TURBO_ENN:
                self._x_tr_list.clear()
                self._y_tr_list.clear()
                self._init_idx = 0
                self._init_lhd = from_unit(
                    latin_hypercube(self._num_init, self._num_dim, rng=self._rng),
                    self._bounds,
                )
                return self._get_init_lhd_points(num_arms)
        if (
            self._mode == TurboMode.TURBO_ONE or self._mode == TurboMode.TURBO_ENN
        ) and self._x_tr_list is not None:
            x_center = self._best_x_tr()
        else:
            x_center = self._best_x()

        def from_unit_fn(x):
            return from_unit(x, self._bounds)

        gp_model = None
        gp_y_mean_fitted = None
        gp_y_std_fitted = None
        weights = None

        if self._mode == TurboMode.TURBO_ONE or self._mode == TurboMode.TURBO_ENN:
            if len(self._x_tr_list) == 0:
                return self._get_init_lhd_points(num_arms)
            if self._trailing_obs is not None:
                x_tr_slice = self._x_tr_list[-self._trailing_obs :]
                y_tr_slice = self._y_tr_list[-self._trailing_obs :]
            else:
                x_tr_slice = self._x_tr_list
                y_tr_slice = self._y_tr_list

            if self._mode == TurboMode.TURBO_ONE:
                import numpy as np

                from .turbo_utils import fit_gp

                gp_model, _likelihood, gp_y_mean_fitted, gp_y_std_fitted = fit_gp(
                    x_tr_slice,
                    y_tr_slice,
                    self._num_dim,
                    num_steps=self._gp_num_steps,
                )
                if gp_model is not None:
                    weights = (
                        gp_model.covar_module.base_kernel.lengthscale.cpu()
                        .detach()
                        .numpy()
                        .ravel()
                    )
                    weights = weights / weights.mean()
                    weights = weights / np.prod(np.power(weights, 1.0 / len(weights)))

        lb_local, ub_local = self._tr_state._compute_bounds_1d(x_center, weights)

        x_cand = raasp(
            x_center,
            lb_local,
            ub_local,
            self._num_candidates,
            num_pert=20,
            rng=self._rng,
            sobol_engine=self._sobol_engine,
        )

        def fallback_fn(x, n):
            return select_uniform(x, n, self._num_dim, self._rng, from_unit_fn)

        if self._mode == TurboMode.TURBO_ZERO:
            return select_uniform(
                x_cand,
                num_arms,
                self._num_dim,
                self._rng,
                from_unit_fn,
            )
        if self._mode == TurboMode.TURBO_ONE:
            selected, self._gp_y_mean, self._gp_y_std, _ = select_gp_thompson(
                x_cand,
                num_arms,
                x_tr_slice,
                y_tr_slice,
                self._num_dim,
                self._gp_num_steps,
                self._rng,
                self._gp_y_mean,
                self._gp_y_std,
                fallback_fn,
                from_unit_fn,
                model=gp_model,
                new_gp_y_mean=gp_y_mean_fitted,
                new_gp_y_std=gp_y_std_fitted,
            )
            return selected
        if self._mode == TurboMode.TURBO_ENN:
            return select_enn_pareto(
                x_cand,
                num_arms,
                x_tr_slice,
                y_tr_slice,
                self._k,
                self._var_scale,
                self._hnsw_threshold,
                self._rng,
                fallback_fn,
                from_unit_fn,
            )
        raise RuntimeError(self._mode)

    def _append_observations(self, x, y) -> None:
        import numpy as np

        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
        if x.ndim != 2 or x.shape[1] != self._num_dim:
            raise ValueError(x.shape)
        if y.ndim != 1 or y.shape[0] != x.shape[0]:
            raise ValueError((x.shape, y.shape))
        if x.shape[0] == 0:
            return
        x_unit = to_unit(x, self._bounds)
        self._x_obs_list.extend(x_unit.tolist())
        self._y_obs_list.extend(y.tolist())
        if self._x_tr_list is not None:
            self._x_tr_list.extend(x_unit.tolist())
            self._y_tr_list.extend(y.tolist())
        y_obs_array = np.asarray(self._y_obs_list, dtype=float)
        self._tr_state.update(y_obs_array)

    def tell(self, x, y) -> None:
        self._append_observations(x, y)

    def _best_x_from_lists(self, x_list, y_list, error_msg: str) -> np.ndarray:
        import numpy as np

        y_array = np.asarray(y_list, dtype=float)
        if y_array.size == 0:
            raise RuntimeError(error_msg)
        idx = argmax_random_tie(y_array, rng=self._rng)
        x_array = np.asarray(x_list, dtype=float)
        return x_array[idx]

    def _draw_initial(self, num_arms: int) -> np.ndarray:
        unit = latin_hypercube(num_arms, self._num_dim, rng=self._rng)
        return from_unit(unit, self._bounds)

    def _get_init_lhd_points(self, num_arms: int, fallback_fn=None) -> np.ndarray:
        import numpy as np

        remaining_init = self._num_init - self._init_idx
        num_to_return = min(num_arms, remaining_init)
        result = self._init_lhd[self._init_idx : self._init_idx + num_to_return]
        self._init_idx += num_to_return
        if num_to_return < num_arms:
            remaining = num_arms - num_to_return
            if fallback_fn is not None:
                result = np.vstack([result, fallback_fn(remaining)])
            else:
                result = np.vstack([result, self._draw_initial(remaining)])
        return result

    def _best_x(self) -> np.ndarray:
        return self._best_x_from_lists(
            self._x_obs_list, self._y_obs_list, "no observations"
        )

    def _best_x_tr(self) -> np.ndarray:
        return self._best_x_from_lists(
            self._x_tr_list, self._y_tr_list, "no trust-region observations"
        )
